"""
Set up the functions that handle how we walk the parsed tree.

Author: Steve Losh (still.life@gmail.com)
"""

import sys
import string

class InputStream(object):
    def __init__(self, stream):
        self.stream = stream
        self.buffer = ""
    
    def read_letter(self):
        if len(self.buffer) == 0:
            letter = self.stream.read(1)
            return letter
        else:
            letter = self.buffer[0]
            self.buffer = self.buffer[1:]
            return letter
    
    def read_word(self):
        word = ""
        
        # Eat all of the whitespace
        letter = self.read_letter()
        while letter in string.whitespace:
            letter = self.read_letter()
        
        # Put all of the non-whitespace characters into the word.
        while letter not in string.whitespace:
            word = word + letter
            letter = self.read_letter()
        
        # Put the last whitespace character back.
        self.buffer = letter + self.buffer
        
        return word
    
    def read_line(self):
        if "\n" in self.buffer:
            line = self.buffer.splitlines()[0]
            self.buffer = "\n".join(self.buffer.splitlines()[1:])
            return line
        else:
            line = self.buffer + self.stream.readline()
            self.buffer = ""
            return line
    

standard_input = InputStream(sys.stdin)

class BreakException(Exception):
    pass

class UndeclaredVariableException(Exception):
    pass

class AlreadyDeclaredVariableException(Exception):
    pass

class SymbolTable:
    def __init__(self, parent):
        self.table = {}
        self.parent = parent
    
    def declareSymbol(self, identifier):
        try:
            self.table[identifier]
        except KeyError, e:
            self.table[identifier] = None
        else:
            raise AlreadyDeclaredVariableException
        
    
    def assignSymbol(self, identifier, value):
        try:
            self.table[identifier]
        except KeyError, e:
            raise UndeclaredVariableException
        
        self.table[identifier] = value
    
    def getSymbol(self, identifier):
        try:
            self.table[identifier]
        except KeyError, e:
            raise UndeclaredVariableException
        
        return self.table[identifier]
    
    def reset(self):
        self.table = {}
    
    def __str__(self):
        return str(self.table)

global_symbols = SymbolTable(None)

def declareVariable(identifier):
    try:
        global_symbols.declareSymbol(identifier)
    except AlreadyDeclaredVariableException, e:
        print "Variable %s is already declared!" % identifier
        raise e

def assignVariable(identifier, value):
    value = evaluate(value)
    try:
        global_symbols.assignSymbol(identifier, value)
    except UndeclaredVariableException, e:
        print "Variable %s is not declared!" % identifier
        raise e

def getVariable(identifier):
    try:
        return global_symbols.getSymbol(identifier)
    except UndeclaredVariableException, e:
        print "Variable %s is not declared!" % identifier
        raise e


def evaluate(Node):
    return Node[0](*Node[1:])


def block(statements):
    for statement in statements:
        evaluate(statement)

def output(print_list, include_newline):
    if include_newline:
        output = []
        for item in print_list:
            output.append(str(evaluate(item)))
        sys.stdout.write(" ".join(output) + "\n")
    else:
        output = []
        for item in print_list:
            output.append(str(evaluate(item)))
        sys.stdout.write(" ".join(output))

def number(n):
    return int(n)

def string_literal(s):
    return s

def loop(identifier, statements):
    done = False
    try:
        while not done:
            for statement in statements:
                evaluate(statement)
    except BreakException:
        pass

def test_equals(a, b):
    return evaluate(a) == evaluate(b)

def test_greater(a, b):
    return evaluate(a) > evaluate(b)

def test_less(a, b):
    return evaluate(a) < evaluate(b)

def test_not(a):
    return not evaluate(a)

def cond(tests, then_statements, else_statements):
    if evaluate(tests):
        for statement in then_statements:
            evaluate(statement)
    else:
        for statement in else_statements:
            evaluate(statement)

def break_loop():
    raise BreakException()

def system_exit(code, message):
    if message != None:
        sys.stderr.write(evaluate(message) + "\n")
    sys.exit(evaluate(code))

def getInput(identifier, kind):
    if "LINE" in kind:
        i = standard_input.read_line()
        if i.endswith('\n'):
            i = i[:-1]
    elif "WORD" in kind:
        i = standard_input.read_word()
    elif "LETTAR" in kind:
        i = standard_input.read_letter()
    
    value = (string_literal, i)
    assignVariable(identifier, value)

def add(a, b):
    return evaluate(a) + evaluate(b)

def sub(a, b):
    return evaluate(a) - evaluate(b)

def mul(a, b):
    return evaluate(a) * evaluate(b)

def div(a, b):
    return evaluate(a) / evaluate(b)

def logical_and(a, b):
    return evaluate(a) and evaluate(b)

def logical_or(a, b):
    return evaluate(a) or evaluate(b)

def logical_xor(a, b):
    if evaluate(a):
        if not evaluate(b):
            return True
        else:
            return False
    else:
        if evaluate(b):
            return True
        else:
            return False


# eof