'''
The lexing and parsing functionality.

Author: Steve Losh (still.life@gmail.com)
'''

from sjlol_func import *
import ply.yacc as yacc
import ply.lex as lex


norm = { "I HAS A": "I_HAS_A",
         "IM IN YR": "IM_IN_YR",
         "IM OUTTA YR": "IM_OUTTA_YR",
         "BIGR THAN": "BIGR_THAN",
         "SMALR THAN": "SMALR_THAN", 
         "?": " ? ",
         "!": " ! ",
         ".": " . ", }

def normalize(s):
    """
    Replace multi-word tokens in the input with their underscored 
    counterparts so the lexer won't get confused.
    """
    for phrase in norm:
        s = s.replace(phrase, norm[phrase])
    return s

def denormalize(s):
    """
    Undo the changes made by normalization.  This is used to fix string
    literals that have been normalizd with the rest of the code.
    """
    for phrase in norm:
        s = s.replace(norm[phrase], phrase)
    return s


# Set up the lexing stuff.
tokens = ( "STARTBLOCK", "ENDBLOCK", "PRINT", "STRING", "DECLARE", "ID",
           "NUMBER", "STARTLOOP", "END", "IF", "NOT", "AND", "OR", "XOR", 
           "GREATER", "LESS", "EQUALS", "EXIT", "BREAK", "THEN", "ELSE", 
           "ASSIGN", "FROM", "INPUT", "EN", "WORD", "LINE", "LETTER", 
           "MUTATEADD", "MUTATESUB", "MUTATEMUL", "MUTATEDIV", 
           "ADD", "SUB", "MUL", "DIV", )
literals = ['"', "!"]
t_ignore = " \t.?"

def t_COMMENT(t):
    r'BTW.*|CAN\ HAS.*|NO\ MOAR'
    pass

def t_NUMBER(t):
    r'\d+'
    try:
         t.value = int(t.value)
    except ValueError:
         print "Number %s is too large!" % t.value
	 t.value = 0
    return t

def t_STRING(t):
    r'"([^"]+|\\"|\\\\)*"'  # I think this is right ...
    t.value = denormalize(t.value[1:-1].decode("string-escape"))
    return t

reserved = {
    "HAI" : "STARTBLOCK",
    "KTHXBYE" : "ENDBLOCK",
    "VISIBLE" : "PRINT",
    "I_HAS_A" : "DECLARE",
    "UP": "INCREMENT",
    "!!": "BANGS",
    "IM_IN_YR" : "STARTLOOP", 
    "IZ": "IF",
    "BIGGER_THAN" : "GREATER",
    "SMALLER_THAN" : "LESS",
    "BYES": "EXIT",
    "DIAF": "EXIT",
    "GTFO": "BREAK",
    "KTHX": "END",
    "YARLY": "THEN",
    "NOWAI": "ELSE",
    "LIEK": "EQUALS",
    "LOL": "ASSIGN",
    "R": "FROM",
    "GIMMEH": "INPUT",
    "NOT": "NOT",
    "N": "EN",
    "LINE": "LINE",
    "WORD": "WORD",
    "LETTAR": "LETTER",
    "UPZ": "MUTATEADD",
    "NERFZ": "MUTATESUB",
    "TIEMZD": "MUTATEMUL",
    "OVARZ": "MUTATEDIV",
    "UP": "ADD",
    "NERF": "SUB",
    "TIEMZ": "MUL",
    "OVAR": "DIV",
    "AND": "AND",
    "OR": "OR",
    "XOR": "XOR",
}

def t_ID(t):
    r'!!|[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'ID')    # Check for reserved words
    return t

def t_newline(t):
    r'\n+'
    t.lexer.lineno += t.value.count("\n")

def t_error(t):
    print "Illegal character '%s'" % t.value[0]
    t.lexer.skip(1)


# Build the lexer
lex.lex()

def p_block(p):
    '''block : STARTBLOCK statements ENDBLOCK'''
    p[0] = (block, p[2])

def p_string(p):
    'string : STRING'
    p[0] = (string_literal, p[1])

def p_print(p):
    '''print : PRINT printlist
             | PRINT printlist "!" '''
    if len(p) == 3:
        p[0] = (output, p[2], True)
    else:
        p[0] = (output, p[2], False)

def p_printlist(p):
    '''printlist : expression
                 | expression EN printlist'''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = [p[1]] + p[3]

def p_declare(p):
    'declare : DECLARE ID'
    p[0] = (declareVariable, p[2])

def p_mutate(p):
    '''mutate : mutateop ID "!" "!"
              | mutateop ID "!" "!" number '''
    if len(p) == 5:
        n = (number, 1)
    else:
        n = p[5]
    
    mutate_operations = {"UPZ": add, "NERFZ": sub, "TIEMZD": mul, "OVARZ": div}
    p[0] = (assignVariable, p[2],
                           (mutate_operations[p[1]], (getVariable, p[2]), n))

def p_mutateop(p):
    '''mutateop : MUTATEADD
                | MUTATESUB
                | MUTATEMUL
                | MUTATEDIV '''
    p[0] = p[1]

def p_loop(p):
    '''loop : STARTLOOP ID statements END'''
    p[0] = (loop, p[2], p[3])

def p_break(p):
    '''break : BREAK'''
    p[0] = (break_loop,)

def p_statements(p):
    '''statements : statements statement
                  | statement '''
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[2]]

def p_statement(p):
    '''statement : block
                 | print 
                 | declare 
                 | loop 
                 | if 
                 | exit 
                 | break 
                 | assignment 
                 | input
                 | mutate'''
    p[0] = p[1]

def p_number(p):
    '''number : NUMBER'''
    p[0] = (number, p[1])

def p_identifier(p):
    '''identifier : ID'''
    p[0] = (getVariable, p[1])

def p_atom(p):
    '''atom : number
            | identifier 
            | string '''
    p[0] = p[1]

def p_expression(p):
    '''expression : atom
                  | expression oper atom'''
    
    operations = {"UP": add, "NERF": sub, "TIEMZ": mul, "OVAR": div}
    
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = (operations[p[2]], p[1], p[3])

def p_oper(p):
    '''oper : ADD
            | SUB
            | MUL
            | DIV '''
    p[0] = p[1]

def p_logicop(p):
    '''logicop : AND
               | OR
               | XOR '''
    p[0] = p[1]

def p_tests(p):
    '''tests : test
             | test logicop tests '''
    logical_operations = { "AND": logical_and, 
                           "OR": logical_or,
                           "XOR": logical_xor }
    
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = (logical_operations[p[2]], p[1], p[3])

def p_test(p):
    '''test : expression comparison expression
            | expression NOT comparison expression'''
    comparisons = { "LIEK": test_equals,
                    "BIGR_THAN": test_greater,
                    "SMALR_THAN": test_less }
    
    if len(p) == 4:
        p[0] = (comparisons[p[2]], p[1], p[3])
    elif len(p) == 5:
        p[0] = (test_not, (comparisons[p[2]], p[1], p[4]))

def p_comparison(p):
    '''comparison : GREATER
                  | LESS
                  | EQUALS '''
    p[0] = p[1]

def p_if(p):
    '''if : IF tests statements END
          | IF tests THEN statements END
          | IF tests THEN statements ELSE statements END'''
    if len(p) < 6:
        then_statements = p[3]
    else:
        then_statements = p[4]
    
    if len(p) < 8:
        else_statements = []
    else:
        else_statements = p[6]
    
    p[0] = (cond, p[2], then_statements, else_statements)

def p_exit(p):
    '''exit : EXIT
            | EXIT number
            | EXIT number string'''
    if len(p) < 4:
        message = None
    else:
        message = p[3]
    
    if len(p) < 3:
        code = (number, 0)
    else:
        code = p[2]
    
    p[0] = (system_exit, code, message)

def p_assignment(p):
    '''assignment : ASSIGN ID FROM expression'''
    p[0] = (assignVariable, p[2], p[4])

def p_input(p):
    '''input : INPUT ID
             | INPUT inputkind ID'''
    if len(p) == 3:
        p[0] = (getInput, p[2], "LINE")
    else:
        p[0] = (getInput, p[3], p[2])

def p_inputkind(p):
    '''inputkind : LINE
                 | WORD
                 | LETTER '''
    p[0] = p[1]

def p_error(p):
    # Error handling is pretty much nonexistant.
    # That's alright though, because lolcats never make mistakes.
    print "Syntax error at token", p.value, "on line", p.lineno


yacc.yacc(debug=0)

def parse(s):
    """
    Normalize and parse the given input string and return the parse tree.
    
    This does NOT evaluate the tree!
    """
    tree = yacc.parse(normalize(s))
    
    return tree


# eof